# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility module for interacting with Matter devices via Pigweed RPC."""
import importlib
import logging
import os
from typing import Callable, Generic, TypeVar

from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import matter_endpoints_base
from gazoo_device.capabilities.interfaces import switchboard_base
from gazoo_device.protos import descriptor_service_pb2
from gazoo_device.switchboard.transports import pigweed_rpc_transport

_ProtobufType = TypeVar("_ProtobufType")
_DESCRIPTOR_SERVICE_NAME = "Descriptor"
_DESCRIPTOR_DEVICE_TYPE_RPC_NAME = "DeviceTypeList"
_DETECT_RETRY = 2
_RPC_TIMEOUT_SEC = 1

logger = gdm_logger.get_logger()


class PigweedProtoState(Generic[_ProtobufType]):
  """Pigweed proto state interface.

  The RPC can accept proto type objects as parameters, however, protobufs aren't
  picklable, so we'll need this class interface to store the encoded protobuf
  object (bytes) and decode it back to the protobuf type.
  """

  def __init__(self, proto_inst: _ProtobufType, decoder_path: str) -> None:
    """Constructor of the PigweedProtoState.

    Args:
      proto_inst: Protobuf instance.
        (ex: device_service_pb2.DeviceInfo(vendor_id=0, product_id=0))
      decoder_path: The import path of the protobuf class.
        (ex: gazoo_device.protos.device_service_pb2.DeviceInfo)
    """
    self._bytes = proto_inst.SerializeToString()
    self._decoder_path = decoder_path

  def _get_decoder(self) -> Callable[[bytes], _ProtobufType]:
    """Gets decoder method from decoder path."""
    module, proto_class = self._decoder_path.rsplit(".", 1)
    mod = importlib.import_module(module)
    return getattr(mod, proto_class).FromString

  def decode(self) -> _ProtobufType:
    """Decodes the encoded proto instance."""
    decoder = self._get_decoder()
    return decoder(self._bytes)


def is_matter_device(
    address: str,
    log_path: str,
    create_switchboard_func: Callable[..., switchboard_base.SwitchboardBase],
    detect_logger: logging.Logger
) -> bool:
  """Returns True if the device is a Matter device.

  Args:
    address: Device serial address.
    log_path: Device log path.
    create_switchboard_func: Method to create the switchboard.
    detect_logger: The logger of device interactions.

  Returns:
    True if the device is a Matter device, False otherwise.
  """
  switchboard = create_switchboard_func(communication_address=address,
                                        communication_type="PigweedSerialComms",
                                        device_name=os.path.basename(log_path),
                                        log_path=log_path,
                                        protobufs=(descriptor_service_pb2,))
  rpc_method = pigweed_rpc_transport.PigweedRPCTransport.rpc
  method_args = (_DESCRIPTOR_SERVICE_NAME, _DESCRIPTOR_DEVICE_TYPE_RPC_NAME)
  method_kwargs = {
      "endpoint": matter_endpoints_base.ROOT_NODE_ENDPOINT_ID,
      "pw_rpc_timeout_s": _RPC_TIMEOUT_SEC}
  is_matter = False
  try:
    # Retry is to avoid flakiness of the descriptor cluster on Matter devices.
    for _ in range(_DETECT_RETRY):
      try:
        ack, _ = switchboard.call(method=rpc_method,
                                  method_args=method_args,
                                  method_kwargs=method_kwargs)
        if ack:
          is_matter = True
          break
      except errors.DeviceError as e:
        detect_logger.info(f"Pigweed RPC failure for address {address}: {e}")
  finally:
    switchboard.close()

  return is_matter



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Class for SSH devices."""
import time

from gazoo_device import config
from gazoo_device import console_config
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.base_classes import gazoo_device_base
from gazoo_device.capabilities import file_transfer_scp
from gazoo_device.capabilities import shell_ssh
from gazoo_device.utility import deprecation_utils
from gazoo_device.utility import host_utils

logger = gdm_logger.get_logger()

# Use a log marker to prevent reading preexisting persistent logs.
_LOG_MARKER = "--- GDM Log Marker ---"
_LOGGING_FILE_PATH = "/var/log/syslog"
_LOG_MARKER_LINE_POS_OR_EMPTY = (
    'grep -n -e "{marker}" {file_path} --text | tail -n 1 | cut -d '
    '":" -f 1').format(
        marker=_LOG_MARKER, file_path=_LOGGING_FILE_PATH)
_LOG_MARKER_LINE_POS = (
    "line_num=$({cmd}); [ -z \"($line_num)\" ] && echo 1 || echo"
    " $line_num".format(cmd=_LOG_MARKER_LINE_POS_OR_EMPTY))

INFO_PREFIX = "INFO_"
COMMANDS = {
    "BOOT_UP_COMPLETE": "echo 'gdm hello'",
    "INJECT_LOG_MARKER": "sudo sh -c 'echo \"{marker}\" >> {file_path}'".format(
        marker=_LOG_MARKER, file_path=_LOGGING_FILE_PATH),
    "LOGGING": "tail -F -n +$({cmd}) {file_path}".format(
        cmd=_LOG_MARKER_LINE_POS, file_path=_LOGGING_FILE_PATH),
    "GDM_HELLO": "echo 'gdm hello'",
}
REGEXES = {}
TIMEOUTS = {
    "BOOT_UP": 60,
    "DISCONNECT": 60,
    "PING_TO_SSH_DELAY": 10,
    "SHELL": 10,
    "SHELL_DEVICE_RESPONSIVENESS": 5,
}


class SshDevice(gazoo_device_base.GazooDeviceBase):
  """Base class for SSH devices."""
  COMMUNICATION_TYPE = "SshComms"
  _COMMUNICATION_KWARGS = {
      "args": host_utils.DEFAULT_SSH_OPTIONS,
      "log_cmd": COMMANDS["LOGGING"],
      "key_info": None,
      "username": "root",
  }

  def __init__(self,
               manager,
               device_config,
               log_file_name=None,
               log_directory=None):
    super().__init__(
        manager,
        device_config,
        log_file_name=log_file_name,
        log_directory=log_directory)
    self._commands.update(COMMANDS)
    self._regexes.update(REGEXES)
    self._timeouts.update(TIMEOUTS)

  @decorators.LogDecorator(logger)
  def _after_boot_hook(self):
    """Executes after the device boots up.

    No-op by default. Override in derived classes as necessary.
    Methods as upgrade, factory reset may also reboot the device.
    """
    pass

  @decorators.health_check
  def check_device_responsiveness(self):
    """Checks if the device is responsive on console.

    Raises:
      DeviceNotResponsiveError: if device is not responsive on console.
    """
    cmd, timeout = self.commands["GDM_HELLO"], self.timeouts[
        "SHELL_DEVICE_RESPONSIVENESS"]
    try:
      self.shell(cmd, timeout=timeout)
    except errors.DeviceError as err:
      raise errors.DeviceNotResponsiveError(
          self.name,
          "unable to execute command {!r} on device's shell".format(cmd),
          timeout=timeout,
          details=str(err))

  @classmethod
  def _convert_build_date(cls, build_date):
    """Hook for build date conversion.

    No-op by default. Override if necessary.

    Args:
      build_date: build_date to convert.

    Returns:
      The converted build date.
    """
    return build_date

  @decorators.CapabilityDecorator(file_transfer_scp.FileTransferScp)
  def file_transfer(self):
    """File transfer capability for moving files from and to the device."""
    return self.lazy_init(
        file_transfer_scp.FileTransferScp,
        ip_address_or_fn=self.ip_address,
        device_name=self.name,
        add_log_note_fn=self.switchboard.add_log_note,
        user=self._COMMUNICATION_KWARGS["username"],
        key_info=self._COMMUNICATION_KWARGS["key_info"])

  def get_console_configuration(self) -> console_config.ConsoleConfiguration:
    """Returns the interactive console configuration."""
    return console_config.get_log_response_separate_port_configuration(
        self.switchboard.get_line_identifier())

  @decorators.LogDecorator(logger)
  def get_detection_info(self):
    """Gets the persistent and optional attributes of a device.

    Returns:
      tuple: (persistent properties dict, optional properties dict)
    """
    self._set_optional_props()
    self._set_persistent_properties()
    self._format_persistent_properties()

    # Reset all capabilities which could have been instantiated
    # during get_detection_info.
    self.reset_all_capabilities()

    return self.props["persistent_identifiers"], self.props["options"]

  @decorators.PersistentProperty
  def health_checks(self):
    """Returns list of methods to execute as health checks."""
    return [
        self.check_device_connected, self.check_create_switchboard,
        self.check_device_responsiveness
    ]

  @decorators.PersistentProperty
  def ip_address(self):
    return self.communication_address

  @classmethod
  def is_connected(cls, device_config):
    """Checks whether or not the device is connected to the computer.

    Args:
      device_config (dict): contains "persistent" dict

    Returns:
      bool: True if device is pingable, False otherwise.
    """
    ip_address = device_config["persistent"]["console_port_name"]
    return host_utils.is_pingable(ip_address)

  @decorators.PersistentProperty
  def os(self):
    return "Linux"

  @decorators.LogDecorator(logger)
  def recover(self, error):
    """Attempts to recover device based on the type of error specified.

    Args:
      error(CheckDeviceReadyError): The exception thrown that will be used to
        identify a possible recovery solution to use.

    Raises:
      DeviceError: If device recovery fails while attempting to perform
        recovery steps.
      CheckDeviceReadyError: If there are no recovery steps available for
        the error argument, it will be re-raised directly.
    """
    if isinstance(error, errors.DeviceNotBootupCompleteError):
      logger.info("%s not fully booted up. Waiting for complete bootup.",
                  self.name)
      self.wait_for_bootup_complete()
    else:
      super().recover(error)

  def shell(self,
            command,
            command_name="shell",
            timeout=None,
            port=0,
            searchwindowsize=config.SEARCHWINDOWSIZE,
            include_return_code=False):
    """Sends command and returns response and optionally return code.

    Can try multiple times as connection can sometimes fail.
    See shell_capability init args for setting the number of retry
    attempts.

    Args:
      command (str): Command to send to the device.
      command_name (str): Identifier for command.
      timeout (float): Time in seconds to wait for device to respond.
      port (int): Which port to send on, 0 or 1.
      searchwindowsize (int): Number of the last bytes to look at
      include_return_code (bool): flag indicating return code should be
        returned.

    Raises:
      DeviceError: If communication fails.

    Returns:
      str: If include_return_code is False return the device response to
        the command.
      tuple: If include_return_code is True return the device response and
        return code.
    """
    timeout = timeout or self.timeouts["SHELL"]
    return self.shell_capability.shell(
        command,
        command_name=command_name,
        timeout=timeout,
        port=port,
        searchwindowsize=searchwindowsize,
        include_return_code=include_return_code)

  @decorators.CapabilityDecorator(shell_ssh.ShellSSH)
  def shell_capability(self):
    return self.lazy_init(
        shell_ssh.ShellSSH,
        send_and_expect=self.switchboard.send_and_expect,
        device_name=self.name,
        timeout=self.timeouts["SHELL"],
        tries=2)

  @decorators.LogDecorator(logger)
  def wait_for_bootup_complete(self, timeout=None):
    """Wait until the device finishes booting up and is ready for testing.

    Args:
      timeout (int): max time to wait for the device to finish booting up.

    Raises:
      DeviceNotBootupCompleteError: device failed to finish booting up
        before the timeout.
    """
    timeout = timeout or self.timeouts["BOOT_UP"]
    shell_resp = ""
    is_shell_responsive = False
    err_msg = ("wait_for_bootup_complete failed. "
               "Device hasn't finished booting in {}s.".format(timeout))

    try:
      shell_resp, return_code = self.shell(
          self.commands["BOOT_UP_COMPLETE"],
          include_return_code=True,
          timeout=timeout)
      is_shell_responsive = return_code == 0

    except errors.DeviceError as err:
      reason = "shell error: {!r}".format(err)
      raise errors.DeviceNotBootupCompleteError(
          self.name, err_msg, reason=reason)

    if not is_shell_responsive:
      reason = ("Shell responded with a non-zero return code. "
                "Response: {}, code: {}.".format(shell_resp, return_code))
      raise errors.DeviceNotBootupCompleteError(
          self.name, err_msg, reason=reason)

  def _format_persistent_properties(self):
    """Cleans up the format of certain persistent properties."""
    persistent_dict = self.props["persistent_identifiers"]
    for key, value in persistent_dict.items():
      if value:
        if "mac_address" in key:
          persistent_dict[key] = value.replace(":", "")
        elif "code" in key:
          persistent_dict[key] = value.splitlines()[-1]
        elif key == "build_date":
          persistent_dict[key] = self._convert_build_date(value)

  def _inject_log_marker(self):
    """Adds a log marker to /var/log/syslog to prevent reading preexisting logs.

    Device logs are read starting with the last log marker (if present).
    """
    self.shell(self.commands["INJECT_LOG_MARKER"])

  def _set_optional_props(self):
    """Sets the optional properties for the device during detection."""
    self.props["options"] = {
        "user_account": None,
        "user_name": None,
    }

  def _set_persistent_properties(self):
    """Obtains persistent properties from the device during detection."""
    for key in [
        cmd for cmd in list(self.commands) if cmd.startswith(INFO_PREFIX)
    ]:
      prop = key.lower()[len(INFO_PREFIX):]  # Remove INFO_ from key.
      response = ""
      resp_list = self.shell(self.commands[key]).split("=")
      if resp_list:
        response = resp_list[-1]

      self.props["persistent_identifiers"][prop] = response

  def _verify_reboot(self, disconnect_timeout=None, bootup_timeout=None):
    """Verifies the device reboots correctly.

    Ensures the device goes offline, comes back online, and finishes booting up.

    Args:
      disconnect_timeout (int): max time to wait for device to go offline & come
          back online.
      bootup_timeout (int): max time to wait for device to finish booting up.

    Raises:
      DeviceNotBootupCompleteError: device failed to boot up.
    """
    disconnect_timeout = disconnect_timeout or self.timeouts["DISCONNECT"]
    bootup_timeout = bootup_timeout or self.timeouts["BOOT_UP"]

    start_time = time.time()
    max_disconnect_time = start_time + disconnect_timeout

    last_ping = True
    while time.time() < max_disconnect_time:
      ping = host_utils.is_pingable(self.ip_address)
      if not ping and not last_ping:
        logger.info("{} offline in {}s.".format(self.name,
                                                int(time.time() - start_time)))
        break
      time.sleep(.5)
      last_ping = ping
    else:
      raise errors.DeviceNotBootupCompleteError(
          self.name, "failed to go offline in {}s".format(disconnect_timeout))

    # close ssh transport as the ssh connection is disconnected.
    self.switchboard.close_all_transports()

    time.sleep(5)  # wait for a bit
    start_time = time.time()
    while time.time() < max_disconnect_time:
      if host_utils.is_pingable(self.ip_address):
        logger.info("{} online in {}s".format(self.name,
                                              int(time.time() - start_time)))
        break
      time.sleep(.5)
    else:
      raise errors.DeviceNotBootupCompleteError(
          self.name,
          "failed to come online {}s".format(int(time.time() - start_time)))

    # There's a delay between the device being responsive to ping
    # and being able to open SSH connections
    time.sleep(self.timeouts["PING_TO_SSH_DELAY"])

    # Reopen ssh transport as ssh connection is once more available.
    self.switchboard.open_all_transports()

    self.wait_for_bootup_complete(timeout=bootup_timeout)
    logger.info("{} booted up successfully in {}s.".format(
        self.name, int(time.time() - start_time)))

    self._after_boot_hook()


deprecation_utils.add_deprecated_attributes(
    SshDevice,
    [("send_file_to_device", "file_transfer.send_file_to_device", True),
     ("recv_file_from_device", "file_transfer.recv_file_from_device", True),
     ("has_command", "shell_capability.has_command", True)])



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Matter unsupported endpoint placeholder class."""
from gazoo_device import decorators
from gazoo_device import gdm_logger
from gazoo_device.capabilities.matter_endpoints.interfaces import unsupported_endpoint_base

logger = gdm_logger.get_logger()


class UnsupportedEndpoint(unsupported_endpoint_base.UnsupportedBase):
  """Matter unsupported endpoint placeholder class."""

  @decorators.DynamicProperty
  def name(self) -> str:
    """The name of the endpoint on the device."""
    return f"Unsupported endpoint (device type: {self.device_type_id})"

  @decorators.DynamicProperty
  def device_type_id(self) -> int:
    """The name of the endpoint on the device."""
    return self._device_type_id



"""LED driver capability.

This capability interacts with the DC2200 from Thorlabs.
Currently supports a single primary LED source.
"""
import typing
from typing import Mapping, Optional

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import led_driver_base
from gazoo_device.utility import retry

import immutabledict
import pyvisa


logger = gdm_logger.get_logger()

_CONNECTION_ATTEMPT_TIMEOUT_S = 10

_COMMANDS = immutabledict.immutabledict({
    "SET_MAX_BRIGHTNESS": "SOURCE:IMOD:HIGH {val}",
    "SET_MIN_BRIGHTNESS": "SOURCE:IMOD:LOW {val}",
    "SET_FREQUENCY": "SOURCE:IMOD:FREQ {freq}",
    "SET_MODULATION_FUNCTION": "SOURCE:IMOD:FUNC {func}",
    "SET_LED_OUTPUT_STATE": "OUTPUT {state}",

    "GET_MAX_BRIGHTNESS": "SOURCE:IMOD:HIGH?",
    "GET_MIN_BRIGHTNESS": "SOURCE:IMOD:LOW?",
    "GET_FREQUENCY": "SOURCE:IMOD:FREQ?",
    "GET_MODULATION_FUNCTION": "SOURCE:IMOD:FUNC?",
    "GET_LED_OUTPUT_STATE": "OUTPUT?",

    "CLEAR_STATUS": "*CLS",
    "ENABLE_REMOTE": "SYSTEM:REMOTE",
})

_MOD_FUNC_SHORT_TO_LONG: Mapping[str, led_driver_base.ModFunc] = (
    immutabledict.immutabledict({
        "SQU": "square",
        "SIN": "sinusoid",
        "TRI": "triangle",
    }))


class LedDriverDefault(led_driver_base.LedDriverBase):
  """Class for LED driver capability."""

  def __init__(
      self,
      device_name: str,
      serial_number: str) -> None:
    """Initializes an instance of the LED driver.

    Args:
      device_name: Name of device. Usually accessible via self.name.
      serial_number: Serial number of the LED driver. This is used
        to identify the power supply and interface with it.
    """
    super().__init__(device_name=device_name)
    self._serial_number = serial_number

    # Note this doesn't hold on to any resources at this stage.
    self._resource_manager = pyvisa.ResourceManager()
    self._resource_string = self._resource_manager.list_resources(
        query=f"?*::{self._serial_number}::?*INSTR")[0]

    self._pyvisa_device_cache = None  # Gets set in health_check.

  def _setup_device(self) -> None:
    """Sets up device by clearing status and enabling remote control."""
    self._pyvisa_device_cache.write(_COMMANDS["CLEAR_STATUS"])
    self._pyvisa_device_cache.write(_COMMANDS["ENABLE_REMOTE"])

  @decorators.DynamicProperty
  def _pyvisa_device(self) -> Optional[pyvisa.resources.resource.Resource]:
    self.health_check()
    return self._pyvisa_device_cache

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def health_check(self) -> None:
    """Checks that the capability is ready to use.

    Raises:
      errors.CapabilityNotReadyError: If failed to connect after
        retrying for _CONNECTION_ATTEMPT_TIMEOUT_S seconds.
    """
    if not self._healthy:
      try:
        # The DC2200 can take a few attempts to open the resource.
        self._pyvisa_device_cache = retry.retry(
            func=self._resource_manager.open_resource,
            func_args=(self._resource_string,),
            exc_type=errors.DeviceError,
            reraise=False,
            timeout=_CONNECTION_ATTEMPT_TIMEOUT_S)
        self._setup_device()
        self._healthy = True
      except Exception as err:
        self._healthy = False
        raise errors.CapabilityNotReadyError(
            msg=repr(err), device_name=self._device_name) from err

  def get_frequency(self) -> float:
    """Gets the LED frequency in Hertz."""
    return float(self._pyvisa_device.query(_COMMANDS["GET_FREQUENCY"]))

  @decorators.CapabilityLogDecorator(logger)
  def set_frequency(self, freq_hz: float) -> None:
    """Sets the LED frequency.

    Args:
      freq_hz: Frequency to set (in Hertz). Min frequency for DC2200 is 20Hz.
    """
    self._pyvisa_device.write(_COMMANDS["SET_FREQUENCY"].format(freq=freq_hz))

  def get_max_brightness(self) -> float:
    """Gets max brightness used for output in percentage 0-100."""
    return float(self._pyvisa_device.query(_COMMANDS["GET_MAX_BRIGHTNESS"]))

  @decorators.CapabilityLogDecorator(logger)
  def set_max_brightness(self, brightness: float) -> None:
    """Sets max brightness used for output in percentage 0-100."""
    self._pyvisa_device.write(
        _COMMANDS["SET_MAX_BRIGHTNESS"].format(val=brightness))

  def get_min_brightness(self) -> float:
    """Gets min brightness used for output in percentage 0-100."""
    return float(self._pyvisa_device.query(_COMMANDS["GET_MIN_BRIGHTNESS"]))

  @decorators.CapabilityLogDecorator(logger)
  def set_min_brightness(self, brightness: float) -> None:
    """Sets min brightness used for output in percentage 0-100."""
    self._pyvisa_device.write(
        _COMMANDS["SET_MIN_BRIGHTNESS"].format(val=brightness))

  def get_modulation_function(self) -> led_driver_base.ModFunc:
    """Gets the modulation function name used by LED driver.

    Raises:
      ValueError: If unsupported function name provided.

    Returns:
      Name of modulation function.
        One of ["sinusoid", "square", "triangle"]
    """
    response = self._pyvisa_device.query(
        _COMMANDS["GET_MODULATION_FUNCTION"]).strip().upper()
    return typing.cast(
        led_driver_base.ModFunc, _MOD_FUNC_SHORT_TO_LONG[response])

  @decorators.CapabilityLogDecorator(logger)
  def set_modulation_function(self, func_name: led_driver_base.ModFunc) -> None:
    """Sets the modulation function name used by LED driver.

    Args:
      func_name: Name of modulation function to use.
        One of ["sinusoid", "square", "triangle"].

    Raises:
      errors.DeviceError: If function name provided is invalid.
    """
    if func_name not in _MOD_FUNC_SHORT_TO_LONG.values():
      raise errors.DeviceError(
          f"{func_name} is not a valid modulation function. "
          f"Please use one of {list(_MOD_FUNC_SHORT_TO_LONG.values())}")
    self._pyvisa_device.write(_COMMANDS["SET_MODULATION_FUNCTION"].format(
        func=func_name))

  def get_output_state(self) -> bool:
    """Gets the LED state boolean - True means on and False means off."""
    return bool(
        int(self._pyvisa_device.query(_COMMANDS["GET_LED_OUTPUT_STATE"])))

  @decorators.CapabilityLogDecorator(logger)
  def set_output_state(self, state: bool) -> None:
    """Sets the LED state boolean - True means on and False means off."""
    self._pyvisa_device.write(
        _COMMANDS["SET_LED_OUTPUT_STATE"].format(state=int(state)))



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ExpectResponse object carries results from calls to GDM expect API.

GDM expect methods, found in switchboard.py, load results into ExpectResponse
to permit calling routines a way to evaluate device execution and behavior.
"""
import dataclasses
import re
from typing import List, Optional


@dataclasses.dataclass
class ExpectResponse(object):
  """Stores the values of expect Attributes.

  Attributes:
      index: index of matching pattern in pattern list (None if timeout).
      before: all the characters looked at before the match.
      after: all the characters after the first matching character.
      match: pattern match object.
      timedout: True if the expect timed out.
      time_elapsed: number of seconds between start and finish.
      remaining: patterns that have not been matched (after a timeout).
      match_list: of re.search Match objects.
  """
  index: Optional[int] = None
  before: str = ""
  after: str = ""
  match: Optional[re.Match] = None
  time_elapsed: float = 0
  timedout: bool = False
  remaining: List[str] = dataclasses.field(default_factory=list)
  match_list: List[re.Match] = dataclasses.field(default_factory=list)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for the Matter cluster capability."""
from typing import Any, Callable
from gazoo_device.capabilities.interfaces import capability_base


class ClusterBase(capability_base.CapabilityBase):
  """Matter cluster capability interface."""

  # Cluster ID defined in the Matter spec.
  CLUSTER_ID = None

  def __init__(self,
               device_name: str,
               endpoint_id: int,
               read: Callable[..., Any],
               write: Callable[..., Any]):
    """Initializes an instance of the Matter cluster capability.

    Args:
      device_name: Device name used for logging.
      endpoint_id: Endpoint ID of endpoint which hosts this cluster.
      read: The Ember API read method.
      write: The Ember API write method.
    """
    super().__init__(device_name=device_name)
    self._endpoint_id = endpoint_id
    self._read = read
    self._write = write



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: device_service.proto
# pylint: skip-file
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gazoo_device.protos import common_pb2 as common__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='device_service.proto',
  package='chip.rpc',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x14\x64\x65vice_service.proto\x12\x08\x63hip.rpc\x1a\x0c\x63ommon.proto\"X\n\x0bPairingInfo\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\x15\n\rdiscriminator\x18\x02 \x01(\r\x12\x0f\n\x07qr_code\x18\x03 \x01(\t\x12\x13\n\x0bqr_code_url\x18\x04 \x01(\t\"\x91\x01\n\nDeviceInfo\x12\x11\n\tvendor_id\x18\x01 \x01(\r\x12\x12\n\nproduct_id\x18\x02 \x01(\r\x12\x18\n\x10software_version\x18\x03 \x01(\r\x12\x15\n\rserial_number\x18\x04 \x01(\t\x12+\n\x0cpairing_info\x18\x05 \x01(\x0b\x32\x15.chip.rpc.PairingInfo\"0\n\nFabricInfo\x12\x11\n\tfabric_id\x18\x01 \x01(\r\x12\x0f\n\x07node_id\x18\x02 \x01(\x04\"X\n\x0b\x44\x65viceState\x12\x1e\n\x16time_since_boot_millis\x18\x01 \x01(\x04\x12)\n\x0b\x66\x61\x62ric_info\x18\x02 \x03(\x0b\x32\x14.chip.rpc.FabricInfo\"\'\n\x0cPairingState\x12\x17\n\x0fpairing_enabled\x18\x01 \x01(\x08\x32\xeb\x03\n\x06\x44\x65vice\x12\x38\n\x0c\x46\x61\x63toryReset\x12\x12.pw.protobuf.Empty\x1a\x12.pw.protobuf.Empty\"\x00\x12\x32\n\x06Reboot\x12\x12.pw.protobuf.Empty\x1a\x12.pw.protobuf.Empty\"\x00\x12\x36\n\nTriggerOta\x12\x12.pw.protobuf.Empty\x1a\x12.pw.protobuf.Empty\"\x00\x12;\n\rGetDeviceInfo\x12\x12.pw.protobuf.Empty\x1a\x14.chip.rpc.DeviceInfo\"\x00\x12=\n\x0eGetDeviceState\x12\x12.pw.protobuf.Empty\x1a\x15.chip.rpc.DeviceState\"\x00\x12?\n\x0fSetPairingState\x12\x16.chip.rpc.PairingState\x1a\x12.pw.protobuf.Empty\"\x00\x12?\n\x0fGetPairingState\x12\x12.pw.protobuf.Empty\x1a\x16.chip.rpc.PairingState\"\x00\x12=\n\x0eSetPairingInfo\x12\x15.chip.rpc.PairingInfo\x1a\x12.pw.protobuf.Empty\"\x00\x62\x06proto3'
  ,
  dependencies=[common__pb2.DESCRIPTOR,])




_PAIRINGINFO = _descriptor.Descriptor(
  name='PairingInfo',
  full_name='chip.rpc.PairingInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='chip.rpc.PairingInfo.code', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='discriminator', full_name='chip.rpc.PairingInfo.discriminator', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='qr_code', full_name='chip.rpc.PairingInfo.qr_code', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='qr_code_url', full_name='chip.rpc.PairingInfo.qr_code_url', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=48,
  serialized_end=136,
)


_DEVICEINFO = _descriptor.Descriptor(
  name='DeviceInfo',
  full_name='chip.rpc.DeviceInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='vendor_id', full_name='chip.rpc.DeviceInfo.vendor_id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='product_id', full_name='chip.rpc.DeviceInfo.product_id', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='software_version', full_name='chip.rpc.DeviceInfo.software_version', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='serial_number', full_name='chip.rpc.DeviceInfo.serial_number', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pairing_info', full_name='chip.rpc.DeviceInfo.pairing_info', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=139,
  serialized_end=284,
)


_FABRICINFO = _descriptor.Descriptor(
  name='FabricInfo',
  full_name='chip.rpc.FabricInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='fabric_id', full_name='chip.rpc.FabricInfo.fabric_id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='node_id', full_name='chip.rpc.FabricInfo.node_id', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=286,
  serialized_end=334,
)


_DEVICESTATE = _descriptor.Descriptor(
  name='DeviceState',
  full_name='chip.rpc.DeviceState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='time_since_boot_millis', full_name='chip.rpc.DeviceState.time_since_boot_millis', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fabric_info', full_name='chip.rpc.DeviceState.fabric_info', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=336,
  serialized_end=424,
)


_PAIRINGSTATE = _descriptor.Descriptor(
  name='PairingState',
  full_name='chip.rpc.PairingState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='pairing_enabled', full_name='chip.rpc.PairingState.pairing_enabled', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=426,
  serialized_end=465,
)

_DEVICEINFO.fields_by_name['pairing_info'].message_type = _PAIRINGINFO
_DEVICESTATE.fields_by_name['fabric_info'].message_type = _FABRICINFO
DESCRIPTOR.message_types_by_name['PairingInfo'] = _PAIRINGINFO
DESCRIPTOR.message_types_by_name['DeviceInfo'] = _DEVICEINFO
DESCRIPTOR.message_types_by_name['FabricInfo'] = _FABRICINFO
DESCRIPTOR.message_types_by_name['DeviceState'] = _DEVICESTATE
DESCRIPTOR.message_types_by_name['PairingState'] = _PAIRINGSTATE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PairingInfo = _reflection.GeneratedProtocolMessageType('PairingInfo', (_message.Message,), {
  'DESCRIPTOR' : _PAIRINGINFO,
  '__module__' : 'device_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.PairingInfo)
  })
_sym_db.RegisterMessage(PairingInfo)

DeviceInfo = _reflection.GeneratedProtocolMessageType('DeviceInfo', (_message.Message,), {
  'DESCRIPTOR' : _DEVICEINFO,
  '__module__' : 'device_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.DeviceInfo)
  })
_sym_db.RegisterMessage(DeviceInfo)

FabricInfo = _reflection.GeneratedProtocolMessageType('FabricInfo', (_message.Message,), {
  'DESCRIPTOR' : _FABRICINFO,
  '__module__' : 'device_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.FabricInfo)
  })
_sym_db.RegisterMessage(FabricInfo)

DeviceState = _reflection.GeneratedProtocolMessageType('DeviceState', (_message.Message,), {
  'DESCRIPTOR' : _DEVICESTATE,
  '__module__' : 'device_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.DeviceState)
  })
_sym_db.RegisterMessage(DeviceState)

PairingState = _reflection.GeneratedProtocolMessageType('PairingState', (_message.Message,), {
  'DESCRIPTOR' : _PAIRINGSTATE,
  '__module__' : 'device_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.PairingState)
  })
_sym_db.RegisterMessage(PairingState)



_DEVICE = _descriptor.ServiceDescriptor(
  name='Device',
  full_name='chip.rpc.Device',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=468,
  serialized_end=959,
  methods=[
  _descriptor.MethodDescriptor(
    name='FactoryReset',
    full_name='chip.rpc.Device.FactoryReset',
    index=0,
    containing_service=None,
    input_type=common__pb2._EMPTY,
    output_type=common__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Reboot',
    full_name='chip.rpc.Device.Reboot',
    index=1,
    containing_service=None,
    input_type=common__pb2._EMPTY,
    output_type=common__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='TriggerOta',
    full_name='chip.rpc.Device.TriggerOta',
    index=2,
    containing_service=None,
    input_type=common__pb2._EMPTY,
    output_type=common__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetDeviceInfo',
    full_name='chip.rpc.Device.GetDeviceInfo',
    index=3,
    containing_service=None,
    input_type=common__pb2._EMPTY,
    output_type=_DEVICEINFO,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetDeviceState',
    full_name='chip.rpc.Device.GetDeviceState',
    index=4,
    containing_service=None,
    input_type=common__pb2._EMPTY,
    output_type=_DEVICESTATE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SetPairingState',
    full_name='chip.rpc.Device.SetPairingState',
    index=5,
    containing_service=None,
    input_type=_PAIRINGSTATE,
    output_type=common__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetPairingState',
    full_name='chip.rpc.Device.GetPairingState',
    index=6,
    containing_service=None,
    input_type=common__pb2._EMPTY,
    output_type=_PAIRINGSTATE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SetPairingInfo',
    full_name='chip.rpc.Device.SetPairingInfo',
    index=7,
    containing_service=None,
    input_type=_PAIRINGINFO,
    output_type=common__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_DEVICE)

DESCRIPTOR.services_by_name['Device'] = _DEVICE

# @@protoc_insertion_point(module_scope)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ADB transport which communicates to the device over a process running 'adb shell'."""
import time

from gazoo_device.switchboard.transports import process_transport
from gazoo_device.utility import adb_utils

READY_RETRY_DELAY = 2


class AdbTransport(process_transport.ProcessTransport):
  """Perform transport communication using the adb executable to some device."""

  def __init__(self,
               comms_address,
               command="shell",
               adb_path=None,
               fastboot_path=None,
               auto_reopen=True,
               open_on_start=True):
    """Initialize the AdbTransport object with the given properties.

    Args:
        comms_address (str): device adb_serial to use when connecting to
          device
        command (str): additional commands to pass to device
        adb_path (str): optional path to adb executable to use
        fastboot_path (str): optional path to fastboot executable to use
        auto_reopen (bool): flag indicating transport should be reopened if
          unexpectedly closed.
        open_on_start (bool): flag indicating transport should be open on
          TransportProcess start.
    """
    self.comms_address = comms_address
    if adb_path is None:
      adb_path = adb_utils.get_adb_path()
    if fastboot_path is None:
      fastboot_path = adb_utils.get_fastboot_path()
    args = "-s {} {}".format(comms_address, command)
    super(AdbTransport, self).__init__(
        command=adb_path,
        args=args,
        auto_reopen=auto_reopen,
        open_on_start=open_on_start)
    self._adb_path = adb_path
    self._fastboot_path = fastboot_path

  def _is_ready_to_open(self):
    result = adb_utils.is_device_online(
        self.comms_address,
        adb_path=self._adb_path,
        fastboot_path=self._fastboot_path)
    if not result:
      time.sleep(READY_RETRY_DELAY)
    return result



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Flashing capability implementation based on Esptool."""

import os
import re
import time
from typing import Callable, Dict, List, NoReturn, Optional

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import flash_build_base
from gazoo_device.capabilities.interfaces import switchboard_base

logger = gdm_logger.get_logger()

# Import path differs for esptool internally and externally.
try:
  # pylint: disable=g-import-not-at-top
  from esptool import esptool
  _ESPTOOL_AVAILABLE = True
except ImportError:
  try:
    # pylint: disable=g-import-not-at-top
    import esptool
    _ESPTOOL_AVAILABLE = True
  except ImportError:
    _ESPTOOL_AVAILABLE = False

_DEFAULT_BOOT_UP_TIMEOUT_SECONDS = 30
_SWITCHBOARD_CAPABILITY = 'switchboard'

# Default flash argumnets used with esptool. For details refer to
# https://docs.espressif.com/projects/esptool/en/latest/esp32/esptool/index.html#esptool-py
_BAUDRATE = 460800
_BEFORE_FLASH = 'default_reset'
_AFTER_FLASH = 'hard_reset'
_FLASH_MODE = 'keep'
_FLASH_FREQ = 'keep'
_FLASH_SIZE = 'detect'
_APPLICATION_OFFSET = 0x20000
_PARTITION_OFFSET = 0x8000
_BOOTLOADER_OFFSET = 0x1000


class FlashBuildEsptool(flash_build_base.FlashBuildBase):
  """Esptool implementation of flash_build capability."""

  def __init__(
      self,
      device_name: str,
      chip_type: str,
      serial_port: str,
      switchboard: Optional[switchboard_base.SwitchboardBase] = None,
      wait_for_bootup_complete_fn: Optional[Callable[[int], None]] = None,
      reset_endpoints_fn: Optional[Callable[[str], None]] = None,
      boot_up_time: int = _DEFAULT_BOOT_UP_TIMEOUT_SECONDS,
      baud: int = _BAUDRATE,
      before: str = _BEFORE_FLASH,
      after: str = _AFTER_FLASH,
      flash_mode: str = _FLASH_MODE,
      flash_freq: str = _FLASH_FREQ,
      flash_size: str = _FLASH_SIZE,
      application_offset: int = _APPLICATION_OFFSET,
      partition_offset: int = _PARTITION_OFFSET,
      bootloader_offset: int = _BOOTLOADER_OFFSET):
    """Initializes an instance of the FlashBuildEsptool capability.

    For details related to before/after flash arguments refer to
      https://docs.espressif.com/projects/esptool/en/latest/esp32/esptool/advanced-options.html#reset-modes
    For details related to flash mode, ferq, size arguments refer to
      https://docs.espressif.com/projects/esptool/en/latest/esp32/esptool/flash-modes.html

    Args:
      device_name: Device name used for logging.
      chip_type: The target device's chip type.
      serial_port: Device serial port.
      switchboard: A Switchboard capability instance if the device supports it.
      wait_for_bootup_complete_fn: The wait_for_bootup_complete method. This
        method will be called after flashing is complete. If not specified,
        time.sleep(boot_up_time) will be used to wait for boot up.
      reset_endpoints_fn: Method to reset matter_endpoint capability.
        This method will be called after flashing is completed.
      boot_up_time: The time to wait for boot up sequence to complete.
      baud: Baudrate for device serial communication.
      before: Action to perform before flashing.
      after: Action to perform after flashing is complete.
      flash_mode: Mode for flashing.
      flash_freq: Clock frequency for SPI flash interactions.
      flash_size: Size of the SPI flash to use.
      application_offset: Memory offset in hex format to flash application
        binary.
      partition_offset: Memory offset in hex format to flash partition-table
        binary.
      bootloader_offset: Memory offset in hex format to flash bootloader binary.

    Raises:
      DependencyUnavailableError: If esptool is not installed.
      ValueError: If chip_type is not supported.
    """
    if not _ESPTOOL_AVAILABLE:
      raise errors.DependencyUnavailableError(
          '"esptool" is not installed. "esptool" is not included due to '
          'licensing restrictions. To enable flashing for this device type, '
          'install "esptool": "pip install esptool>=3.2".')

    if chip_type not in esptool.SUPPORTED_CHIPS:  # pytype: disable=module-attr
      raise ValueError(f'Chip {chip_type} not supported by esptool.')

    super().__init__(device_name=device_name)
    self._serial_port = serial_port
    self._chip_type = chip_type
    self._flash_args = {
        'baud': baud,
        'before': before,
        'after': after,
        'flash_mode': flash_mode,
        'flash_freq': flash_freq,
        'flash_size': flash_size,
        'application_offset': application_offset,
        'partition_offset': partition_offset,
        'bootloader_offset': bootloader_offset,
    }
    self._switchboard = switchboard
    self._wait_for_bootup_complete_fn = wait_for_bootup_complete_fn
    self._reset_endpoints_fn = reset_endpoints_fn
    self._boot_up_time = boot_up_time

  @decorators.CapabilityLogDecorator(logger)
  def _get_write_command_arguments(self, erase: bool = False) -> List[str]:
    """Helper function to get write command arguments.

    For details please refer to
    https://docs.espressif.com/projects/esptool/en/latest/esp32/esptool/scripting.html#embedding-into-custom-scripts

    Args:
      erase: True to use erase-all option when flashing, else False.

    Returns:
      Arguments required for flashing.
    """
    basic_command_args = [
        '--port', self._serial_port, '--baud', self._flash_args['baud'],
        '--chip', self._chip_type, '--before', self._flash_args['before'],
        '--after', self._flash_args['after']
    ]
    write_command_args = [
        'write_flash', '--flash_freq', self._flash_args['flash_freq'],
        '--flash_mode', self._flash_args['flash_mode'], '--flash_size',
        self._flash_args['flash_size'], '--compress'
    ]
    if erase:
      write_command_args.append('--erase-all')
    return basic_command_args + write_command_args

  @decorators.CapabilityLogDecorator(logger)
  def _verify_file(self, image_path: str, ends_with: str = '.bin') -> None:
    """Helper function to validate image file.

    Args:
      image_path: Path to the image file.
      ends_with: Expected image suffix. E.g bootloader files are expected to
        have suffix 'bootloader.bin'.

    Raises:
      ValueError: If the image file suffix does not end with 'ends_with' arg.
      FileNotFoundError: If image file does not exist.
    """
    if not image_path.endswith(ends_with):
      raise ValueError(f'Only {ends_with} type file can be flashed.')
    if not os.path.exists(image_path):
      raise FileNotFoundError(f'Firmware image {image_path} does not exist.')

  @decorators.CapabilityLogDecorator(logger)
  def flash_device(self,
                   list_of_files: List[str],
                   expected_version: Optional[str] = None,
                   expected_build_type: Optional[str] = None,
                   verify_flash: bool = True,
                   method: Optional[str] = None,
                   erase_flash: bool = False) -> None:
    """Flashes the firmware image (.bin file) on the device.

    Args:
      list_of_files: List of image files on local host in following order
        [build_file_name, bootloader_file_name, partition_table_file_name].
      expected_version: Not used.
      expected_build_type: Not used.
      verify_flash: Not used. 'esptool' always verifies flashed image.
      method: Not used.
      erase_flash: True if everything needs to be erased before flashing.

    Raises:
      ValueError: If list of files is empty, all values are None or list length
        is more than 3.
    """
    del expected_version, expected_build_type, verify_flash, method  # Unused.
    if len(list_of_files) != 3:
      raise ValueError('Only application, bootloader, partition table files can'
                       ' be flashed via esptool. If only flashing one type'
                       ' please set others to None.')
    if not any(list_of_files):
      raise ValueError('No firmware files provided to flash the device.')

    command = self._get_write_command_arguments(erase=erase_flash)

    offset_types = [
        'application_offset', 'bootloader_offset', 'partition_offset'
    ]
    ends_with = ['.bin', 'bootloader.bin', 'partition-table.bin']
    for image, offset_type, ends_with in zip(list_of_files, offset_types,
                                             ends_with):
      if image is not None:
        self._verify_file(image, ends_with)
        command += [self._flash_args[offset_type], image]

    command = list(map(str, command))

    # Close serial connection from GDM to avoid conflict with esptool flashing.
    if self._switchboard is not None:
      self._switchboard.close_all_transports()

    logger.info(f'Executing esptool command: {command}')
    esptool.main(command)  # pytype: disable=module-attr

    if self._switchboard is not None:
      self._switchboard.open_all_transports()

    if self._flash_args['after'] == _AFTER_FLASH:
      if self._wait_for_bootup_complete_fn:
        self._wait_for_bootup_complete_fn(self._boot_up_time)
      else:
        time.sleep(self._boot_up_time)

    # For Matter device classes, we'll need to reset the Matter endpoint mapping
    # as the supported endpoints might change after flashing a new build.
    if self._reset_endpoints_fn is not None:
      self._reset_endpoints_fn()

  @decorators.CapabilityLogDecorator(logger)
  def download_build_file(self,
                          remote_build_folder: str,
                          local_folder: str) -> NoReturn:
    """Retrieves the build file(s) from the remote location."""
    del remote_build_folder, local_folder
    raise NotImplementedError(
        'download_build_file is not available in flash_build_esptool for now.')

  @decorators.CapabilityLogDecorator(logger)
  def get_defaults(self) -> NoReturn:
    """Returns a dictionary of default build arguments."""
    raise NotImplementedError(
        'get_defaults is not available in flash_build_esptool for now.')

  @decorators.CapabilityLogDecorator(logger)
  def get_firmware_type(self, build_args: Optional[str] = None) -> str:
    """Returns the firmware type based on the build arguments."""
    del build_args
    return flash_build_base.UNKNOWN

  @decorators.CapabilityLogDecorator(logger)
  def get_firmware_version(self, build_args: Optional[str] = None) -> str:
    """Returns the firmware version based on the build arguments."""
    del build_args
    return flash_build_base.UNKNOWN

  @decorators.CapabilityLogDecorator(logger)
  def get_remote_build_folder(self,
                              build_args: Optional[str] = None) -> NoReturn:
    """Uses the build arguments to determine the remote build folder."""
    raise NotImplementedError('get_remote_build_folder is not available in '
                              'flash_build_esptool for now.')

  @decorators.CapabilityLogDecorator(logger)
  def extract_build_info(self, build_args: Optional[str] = None) -> NoReturn:
    """Converts the provided build arguments into info about the build."""
    raise NotImplementedError(
        'extract_build_info is not available in flash_build_esptool for now.')

  @decorators.CapabilityLogDecorator(logger)
  def latest_verified_build_folder(self) -> NoReturn:
    """Returns the remote build folder path for the latest verified build."""
    raise NotImplementedError(
        'latest_verified_build_folder is not available in '
        'flash_build_esptool for now.')

  @decorators.CapabilityLogDecorator(logger)
  def upgrade(
      self,
      build_number: Optional[int] = None,
      build_url: Optional[str] = None,
      build_file: Optional[str] = None,
      forced_upgrade: bool = False,
      latest_verified: bool = False,
      partition_file: Optional[str] = None,
      bootloader_file: Optional[str] = None,
      erase_flash: bool = False,
      baud: Optional[int] = None,
      before: Optional[str] = None,
      after: Optional[str] = None,
      flash_mode: Optional[str] = None,
      flash_freq: Optional[str] = None,
      flash_size: Optional[str] = None,
      application_offset: Optional[int] = None,
      partition_offset: Optional[int] = None,
      bootloader_offset: Optional[int] = None,
      flash_settings_file: Optional[str] = None,
      **other_build_args: str,
  ) -> None:
    """Upgrade the device based on the provided build arguments.

    For details related to before/after flash arguments refer to
      https://docs.espressif.com/projects/esptool/en/latest/esp32/esptool/advanced-options.html#reset-modes

    For details related to flash mode, ferq, size arguments refer to
      https://docs.espressif.com/projects/esptool/en/latest/esp32/esptool/flash-modes.html#

    Each entry in the partition table file has a name (label), type
    (app, data, or something else), subtype and the offset in flash where the
    partition is loaded. For details refer to
    https://docs.espressif.com/projects/esp-idf/en/latest/esp32/api-guides/partition-tables.html#partition-tables

    A Bootloader selects and loads application image into RAM. To update the
    device bootloder image please specify a bootloader.bin generated when
    compiling a esp32 project. For details refer to
    https://docs.espressif.com/projects/esp-idf/en/latest/esp32/api-guides/bootloader.html#bootloader

    Args:
     build_number: Not used.
     build_url: Not used.
     build_file: Local path to the application file.
     forced_upgrade: Not used.
     latest_verified: Not used.
     partition_file: Local path to the partition table file.
     bootloader_file: Local path to the bootloader file.
     erase_flash: True if everything needs to be erased before flashing.
     baud: Baudrate for device serial communication.
     before: Action to perform before flashing.
     after: Action to perform after flashing is complete.
     flash_mode: Mode for flashing.
     flash_freq: Clock frequency for SPI flash interactions.
     flash_size: Size of the SPI flash to use.
     application_offset: Memory offset in hex format to flash application
       binary.
     partition_offset: Memory offset in hex format to flash partition-table
       binary.
     bootloader_offset: Memory offset in hex format to flash bootloader binary.
     flash_settings_file: *.flash.py present when using a matter sample app
       from matter-automation-project. This file will be used to read default
       flash settings. These can be overridden if user explicitly passes them as
       an argument.
     **other_build_args: Not used.
    """
    del build_number, build_url, forced_upgrade, latest_verified  # Unused.
    del other_build_args  # Unused.
    build_args = {
        name: value for name, value in locals().items() if name not in ['self']
    }
    self.upgrade_over_the_wire(**build_args)

  @decorators.CapabilityLogDecorator(logger)
  def upgrade_over_the_wire(self, **build_args: str) -> None:
    """Using the build arguments, flash the build on the device.

    Args:
      **build_args: Dictionary of build arguments.
    """
    list_of_files = [
        build_args['build_file'], build_args['bootloader_file'],
        build_args['partition_file']
    ]

    flash_args_keys = self._flash_args.keys()

    # Update flash arguments from *.flash.py
    if build_args['flash_settings_file']:
      with open(build_args['flash_settings_file']) as settings_file:
        settings = settings_file.read()
        # Regex for flash arguments such as `'flash_freq': '30m',`.
        pattern = re.compile(r"'(\w+)': '(\w+)',", re.MULTILINE | re.ASCII)
        self._flash_args.update({
            match.group(1): match.group(2)
            for match in re.finditer(pattern, settings)
            if match.group(1) in flash_args_keys and match.group(2) is not None
        })

    # Override flash settings with user provided arguments.
    self._flash_args.update({
        build_args_key: build_args_value
        for build_args_key, build_args_value in build_args.items()
        if build_args_key in flash_args_keys and build_args_value is not None
    })

    self.flash_device(
        list_of_files=list_of_files, erase_flash=build_args['erase_flash'])



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""RPC implementation of Matter Temperature Measurement cluster capability.
"""
from gazoo_device import decorators
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters.interfaces import temperature_measurement_base
from gazoo_device.protos import attributes_service_pb2

_MIN_MEASURED_UPPERBOUND = 32766
_MAX_MEASURED_UPPERBOUND = 32767
_ATTR_VAL_MAX = 1 << 16
_TempMeasurementCluster = matter_enums.TemperatureMeasurementCluster
INT16_ATTRIBUTE_TYPE = attributes_service_pb2.AttributeType.ZCL_INT16S_ATTRIBUTE_TYPE


def _convert_attribute_constraint_value(
    unsigned_value: int, upper_bound: int) -> int:
  """Converts the unsigned value to the attribute constraint value.

  The attribute values of temperature_measurement has constraint according to
  the spec: MinMeasuredValue has range from -27315 to 32766; MaxMeasuredValue
  has range from -27314 to 32767. However, the returned value from the Ember
  API call will always be an unsigned value (which is always >= 0), therefore
  we'll need to manually substract the 2^16 for complement. (ex: returned value
  = 38221 is actually 38221 - 2^16 = -27315 in spec)

  Args:
    unsigned_value: The unsigned value of attribute returned from Ember API.
    upper_bound: Upper bound value of attribute from spec.

  Returns:
    The actual attribute value.
  """
  if unsigned_value <= upper_bound:
    return unsigned_value
  return unsigned_value - _ATTR_VAL_MAX


class TemperatureMeasurementClusterPwRpc(
    temperature_measurement_base.TemperatureMeasurementClusterBase):
  """Matter Temperature Measurement cluster capability."""

  @decorators.DynamicProperty
  def measured_value(self) -> int:
    """The MeasuredValue attribute.

    Represents the temperature in degrees Celsius as follows:
    MeasuredValue = 100 x temperature [°C]
    Where -273.15°C ≤ temperature ≤ 327.67°C, with a resolution of 0.01°C.

    Returns:
      The MeasuredValue attribute.
    """
    measured_value_data = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=_TempMeasurementCluster.ID,
        attribute_id=_TempMeasurementCluster.ATTRIBUTE_MEASURED_VALUE,
        attribute_type=INT16_ATTRIBUTE_TYPE)
    return measured_value_data.data_int16

  @decorators.DynamicProperty
  def min_measured_value(self) -> int:
    """The MinMeasuredValue attribute.

    The MinMeasuredValue attribute indicates the minimum value of MeasuredValue
    that is capable of being measured.

    Returns:
      The MinMeasuredValue attribute.
    """
    min_measured_value_data = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=_TempMeasurementCluster.ID,
        attribute_id=_TempMeasurementCluster.ATTRIBUTE_MIN_MEASURED_VALUE,
        attribute_type=INT16_ATTRIBUTE_TYPE)
    return _convert_attribute_constraint_value(
        min_measured_value_data.data_int16, _MIN_MEASURED_UPPERBOUND)

  @decorators.DynamicProperty
  def max_measured_value(self) -> int:
    """The MaxMeasuredValue attribute.

    The MaxMeasuredValue attribute indicates the maximum value of MeasuredValue
    that is capable of being measured.

    Returns:
      The MaxMeasuredValue attribute.
    """
    max_measured_value_data = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=_TempMeasurementCluster.ID,
        attribute_id=_TempMeasurementCluster.ATTRIBUTE_MAX_MEASURED_VALUE,
        attribute_type=INT16_ATTRIBUTE_TYPE)
    return _convert_attribute_constraint_value(
        max_measured_value_data.data_int16, _MAX_MEASURED_UPPERBOUND)



"""LED driver capability interface."""
import abc

from gazoo_device.capabilities.interfaces import capability_base
from typing_extensions import Literal

ModFunc = Literal["sinusoid", "square", "triangle"]


class LedDriverBase(capability_base.CapabilityBase):
  """Abstract class for LED driver interface."""

  @abc.abstractmethod
  def get_frequency(self) -> float:
    """Gets the LED frequency in Hertz."""

  @abc.abstractmethod
  def set_frequency(self, freq_hz: float) -> None:
    """Sets the LED frequency.

    Args:
      freq_hz: Frequency to set (in Hertz). Min frequency for DC2200 is 20Hz.
    """

  @abc.abstractmethod
  def get_max_brightness(self) -> int:
    """Gets max brightness used for output in percentage 0-100."""

  @abc.abstractmethod
  def set_max_brightness(self, brightness: int) -> None:
    """Sets max brightness used for output in percentage 0-100."""

  @abc.abstractmethod
  def get_min_brightness(self) -> int:
    """Gets min brightness used for output in percentage 0-100."""

  @abc.abstractmethod
  def set_min_brightness(self, brightness: int) -> None:
    """Sets min brightness used for output in percentage 0-100."""

  @abc.abstractmethod
  def get_modulation_function(self) -> ModFunc:
    """Gets the modulation function name used by LED driver.

    Raises:
      ValueError: If unsupported function name provided.

    Returns:
      Name of modulation function.
        One of ["sinusoid", "square", "triangle"]
    """

  @abc.abstractmethod
  def set_modulation_function(self, func_name: ModFunc) -> None:
    """Sets the modulation function name used by LED driver.

    Args:
      func_name: Name of modulation function to use.
        One of ["sinusoid", "square", "triangle"].

    Raises:
      ValueError: If unsupported function name provided.
    """

  @abc.abstractmethod
  def get_output_state(self) -> bool:
    """Gets the LED state boolean - True means on and False means off."""

  @abc.abstractmethod
  def set_output_state(self, state: bool) -> None:
    """Sets the LED state boolean - True means on and False means off."""



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Device controller for EFR32 Matter device.

This device controller populates the supported Matter endpoints on the EFR32
platform by using the descriptor RPC service.
"""
from gazoo_device import decorators
from gazoo_device import detect_criteria
from gazoo_device import gdm_logger
from gazoo_device.base_classes import matter_device_base
from gazoo_device.capabilities import flash_build_jlink

logger = gdm_logger.get_logger()
_EFR32_JLINK_NAME = "EFR32MG12PXXXF1024"


class Efr32Matter(matter_device_base.MatterDeviceBase):
  """EFR32 Matter device controller."""
  DETECT_MATCH_CRITERIA = {
      detect_criteria.PigweedQuery.IS_MATTER: True,
      detect_criteria.PigweedQuery.MANUFACTURER_NAME: r"silicon(_| )labs",
      detect_criteria.PigweedQuery.PRODUCT_NAME: "j-link",
  }
  # Button definition on EFR32 dev board:
  # https://github.com/project-chip/connectedhomeip/tree/master/examples/lighting-app/efr32#running-the-complete-example
  VALID_BUTTON_IDS = (0, 1)

  DEVICE_TYPE = "efr32matter"

  @decorators.PersistentProperty
  def os(self) -> str:
    """OS of EFR32 platform."""
    return "FreeRTOS"

  @decorators.PersistentProperty
  def platform(self) -> str:
    """EFR32 Silabs platform."""
    return "EFR32MG"

  @decorators.CapabilityDecorator(flash_build_jlink.FlashBuildJLink)
  def flash_build(self) -> flash_build_jlink.FlashBuildJLink:
    """FlashBuildJLink capability to flash hex image."""
    return self.lazy_init(flash_build_jlink.FlashBuildJLink,
                          device_name=self.name,
                          serial_number=self.serial_number,
                          platform_name=_EFR32_JLINK_NAME,
                          reset_endpoints_fn=self.matter_endpoints.reset,
                          switchboard=self.switchboard)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TcpTransport is used for connection-oriented transmissions.

TcpTransport is responsible for delivering data to the appropriate application
process
on the host computer. A host and port number are combined to create a network
socket (an
identification address of a process-to-process communication). The socket can
then used by an application to handle a data stream for various needs.
"""
import socket
import six
from gazoo_device.switchboard import transport_properties
from gazoo_device.switchboard.transports import transport_base

CONNECT_TIMEOUT = 10.0  # in secs


def check_port(port):
  """Verifies port value given is valid.

  Args:
      port (int): port number to verify

  Raises:
      ValueError: if port number provided is invalid

  Returns:
      int: port number
  """

  if port < 0 or port > 65535:
    raise ValueError("Port {} out of range".format(port))
  return port


def check_host(host):
  """Verifies host given is valid.

  Args:
      host (str): Either IP Address or hostname to verify

  Raises:
      ValueError: if ip address provided is invalid

  Returns:
      str: host
  """

  try:  # Valid ip address?
    socket.inet_pton(socket.AF_INET, host)
  except socket.error:
    try:  # Valid hostname?
      socket.getaddrinfo(host, None)
    except socket.error:
      raise ValueError("Host Address {} not valid".format(host))
  return host


class TcpTransport(transport_base.TransportBase):
  """A TCP Transport class."""

  def __init__(self,
               comms_address,
               args,
               connect_timeout=CONNECT_TIMEOUT,
               auto_reopen=False,
               open_on_start=False):
    """Initialize the TcpTransport object with the given TCP properties.

    Args:
        comms_address (str): hostname or IP address of the device to connect
          to.
        args (str): The port to communicate to.
        connect_timeout (float): timeout on the socket instance before
          attempting connect.
        auto_reopen (bool): flag indicating transport should be reopened if
          unexpectedly closed.
        open_on_start (bool): flag indicating transport should be open on
          TransportProcess start.

    Raises:
        ValueError: host not valid or port out of range
    """
    self.comms_address = comms_address
    port = int(args)
    super(TcpTransport, self).__init__(auto_reopen, open_on_start)
    self._properties.update({
        transport_properties.HOST: comms_address,
        transport_properties.PORT: int(args),
        transport_properties.CONNECT_TIMEOUT: connect_timeout,
    })
    check_host(comms_address)
    check_port(port)
    self._socket = None

  def is_open(self):
    """Returns True if tcp socket has been created.

    Returns:
        bool: True if _socket is not None
    """
    return hasattr(self, "_socket") and self._socket is not None

  def _open(self):
    """Opens or reopens the tcp connection using the current property values.

    Raises:
        IOError: if host and port do not resolve
    """
    self._socket = socket.create_connection(
        (self._properties[transport_properties.HOST],
         self._properties[transport_properties.PORT]),
        self._properties[transport_properties.CONNECT_TIMEOUT])

  def _close(self):
    """Closes the TCP connection."""
    self._socket.shutdown(socket.SHUT_RDWR)
    self._socket.close()
    self._socket = None

  def _read(self, size=1, timeout=None):
    """Returns bytes read up to max_bytes within timeout in seconds specified.

    Args:
        size (int): maximum number of bytes to read within timeout seconds
        timeout (float): maximum seconds to wait to read bytes or
          indefinitely
                         if timeout=None. Default: None

    Returns:
        str: bytes read from transport
             or "" if no bytes were read
             or "" if timeout was reached
    """
    try:
      self._socket.settimeout(timeout)
      read_bytes = self._socket.recv(size)
      return "" if read_bytes is None else read_bytes
    except socket.timeout:
      return ""

  def _write(self, data, timeout=None):
    """Writes the command provided to the TCP socket.

    Args:
        data (str): bytes to be written out within timeout seconds
        timeout (float): maximum seconds to wait to write bytes or
          indefinitely
                         if timeout=None. Default: None

    Returns:
        int: number of bytes written or None if no bytes were written
    """
    if isinstance(data, six.text_type):
      data = data.encode("utf-8", errors="replace")

    self._socket.settimeout(timeout)
    self._socket.sendall(data)
    return len(data)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common reusable utility functions."""
import re
import weakref


class MethodWeakRef(object):
  """Allows creating weak references to instance methods.

  Note:
      using weakref.ref() on an instance method directly returns a dead
      reference. See https://stackoverflow.com/questions/599430.
  """

  def __init__(self, instance_method):
    """Create a weak reference to an instance (bound) method.

        Args:
            instance_method (method): instance (bound) method.
    """
    self._func = instance_method.__func__
    self._instance_weakref = weakref.ref(instance_method.__self__)

  def __call__(self, *args, **kwargs):
    """Calls the instance method if the instance is still alive."""
    instance = self._instance_weakref()
    if instance is not None:
      self._func(instance, *args, **kwargs)


def generate_name(_object):
  """Generates a snake_case name (str) for the given object.

  Args:
      _object (object): any python object which has a non-empty __name__
        attribute.

  Raises:
      ValueError: _object does not have __name__ attribute, or __name__ is "".

  Returns:
      str: snake_case name for the object.
  """
  object_name = getattr(_object, "__name__", "")
  if not object_name:
    raise ValueError(
        "Object {} must have a non-empty __name__ attribute.".format(_object))

  if "_" in object_name:  # Presumably already snake_case
    return object_name.lower()
  else:  # Presumably TitleCase
    return title_to_snake_case(object_name)


def get_value_from_json(json_data, key_sequence, raise_if_absent=True):
  """Extracts a value from a JSON dictionary via provided series of keys.

  This function allows easy extraction of a value stored in a nested
  dictionary structure, with validation of the existence of the necessary
  keys at each level of nesting.

  Args:
      json_data (dict): the response object from which the value will be
        extracted.
      key_sequence (list): an ordered list of string keys representing the
        path through the dictionary necessary to reach the desired value.
                           Ex. to extract foo['baz']['a'] the call would look
                             like: get_value_from_json(foo, ['baz', 'a'])  and
                             would return 'some text' given the example data
                             above.
      raise_if_absent (bool): if False, return None if a key is not present.
        if True, raise KeyError if a key is not present.

  Returns:
      object: the requested JSON field.

  Raises:
      KeyError: If the provided response message doesn't contain the necessary
      keys/values AND if raise_if_absent is True.
  """
  current_dict = json_data
  key_text = ""
  for key in key_sequence:
    if key in current_dict:
      current_dict = current_dict[key]
    else:
      if raise_if_absent:
        raise KeyError(
            "Unable to find key '{}' in json_data{}: json_data = {}".format(
                key, key_text, json_data))
      else:
        return None
    key_text += "['{}']".format(key)
  return current_dict


def title_to_snake_case(s):
  """Convert TitleCase string to snake_case.

  Args:
      s (str): TitleCase string.

  Returns:
      str: snake_case string.

  Raises:
      ValueError: provided string contains underscores.

  Note:
      consecutive capital characters are supported ("ABc"), but underscores
      are not.
  """
  if "_" in s:
    raise ValueError(
        "{} is not a TitleCase string (found underscores).".format(s))

  word_starts = [pos for pos in range(len(s)) if _is_new_word(s, pos)
                ] + [len(s)]
  words = [
      s[word_starts[idx]:word_starts[idx + 1]].lower()
      for idx in range(len(word_starts) - 1)
  ]
  return "_".join(words)


def _is_new_word(s, pos):
  """Returns whether a new words starts at s[pos] in a TitleCase string."""
  return (pos == 0 or (s[pos].isupper() and
                       (not s[pos - 1].isupper() or
                        (pos + 1 < len(s) and not s[pos + 1].isupper()))))


def extract_posix_portable_characters(string: str) -> str:
  """Extracts posix fully portable characters from a string.

  Args:
    string: String to extract characters from.

  Returns:
    String of characters with all non-posix portable characters removed.
  """
  return re.sub(r"[^a-zA-Z0-9._\-]", "", string)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shell capability interface."""
import abc
from typing import Callable, Optional, Tuple, Union

from gazoo_device import config
from gazoo_device.capabilities.interfaces import capability_base
from gazoo_device.switchboard import expect_response

TIMEOUT = 10
TRIES = 2

_DEVICE_CAPABILITY_NAME = "shell_capability"


class ShellBase(capability_base.CapabilityBase):
  """Abstract base class defining the API for the shell() method."""

  def __init__(self,
               send_and_expect: Callable[..., expect_response.ExpectResponse],
               device_name: str,
               tries: int = TRIES,
               timeout: float = TIMEOUT) -> None:
    """Initializes an instance of Shell capability.

    Args:
      send_and_expect: send_and_expect method of the device class instance.
      device_name: Name of the device using this capability.
      tries: How many times to try sending the shell command.
      timeout: Shell timeout in seconds.
    """
    super().__init__(device_name=device_name)

    self._send_and_expect = send_and_expect
    self._tries = tries
    self._timeout = timeout

  @classmethod
  def get_capability_name(cls) -> str:
    """Override the default to avoid a name conflict with the "shell" method."""
    return _DEVICE_CAPABILITY_NAME

  @abc.abstractmethod
  def shell(
      self,
      command: str,
      command_name: str = "shell",
      timeout: Optional[float] = None,
      port: int = 0,
      include_return_code: bool = False,
      searchwindowsize: int = config.SEARCHWINDOWSIZE
  ) -> Union[str, Tuple[str, int]]:
    """Sends command and returns response and optionally return code.

    If the SSH connection fails, the command is retried (up to a total of
    self._tries attempts).

    Args:
      command: Command to send to the device.
      command_name: Identifier for command.
      timeout: Time in seconds to wait for device to respond.
      port: Which port to send on. Port 0 is typically used for commands.
      include_return_code: Whether to also return the command return code.
      searchwindowsize: Number of the last bytes to look at.

    Raises:
      DeviceError: if communication fails.

    Returns:
      If include_return_code is False, returns the device response to the
        command.
      If include_return_code is True, returns the device response and the
        command return code.
    """

  @abc.abstractmethod
  def has_command(self, binary_name: str) -> bool:
    """Returns if binary_name is installed on the device.

    Args:
      binary_name: Name of the executable.
    """



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes and variables to share across usb_info modules."""
import inspect

ANDROID_NAMES = ("Android", "Pixel")
CAMBRIONIX_NAMES = ("FT230X Basic UART", "PS15-USB3")
CAMBRIONIX_VENDOR_PRODUCT_ID = ("0403/6015", "2cd9/0021")

CAMBRIONIX_PORT_MAP = {
    "PP8S": {
        "1.1": 1,
        "1.2": 2,
        "1.3": 3,
        "1.4": 4,
        "2.1": 8,
        "2.2": 7,
        "2.3": 6,
        "2.4": 5
    },
    "PP15S": {
        "1.1": 4,
        "1.2": 5,
        "1.3": 6,
        "1.4": 7,
        "2.1": 15,
        "2.2": 14,
        "2.3": 13,
        "2.4": 12,
        "3.1": 11,
        "3.2": 10,
        "3.3": 9,
        "3.4": 8,
        "4.2": 1,
        "4.3": 2,
        "4.4": 3
    },
    "PS15-USB3": {
        "1.1": 1,
        "1.2": 2,
        "1.3": 9,
        "1.4": 8,
        "2.1": 6,
        "2.2": 7,
        "2.3": 15,
        "2.4": 14,
        "3.1": 3,
        "3.2": 4,
        "3.3": 11,
        "3.4": 10,
        "4.1": 5,
        "4.3": 13,
        "4.4": 12
    },
    "U16S": {
        "2": 1,
        "5": 2,
        "4.1": 4,
        "4.2": 3,
        "4.3": 5,
        "4.4": 6,
        "4.5": 8,
        "4.6": 7,
        "4.7": 9,
        "7.1": 11,
        "7.2": 13,
        "7.3": 14,
        "7.4": 15,
        "7.5": 16,
        "7.6": 10,
        "7.7": 12
    }
}


class UsbInfo(object):
  """Class for storing all info about usb devices in a system agnostic way."""
  _address = None
  _child_addresses = []
  _ftdi_interface = 0
  _manufacturer = ""
  _disk = None
  _product_id = ""
  _product_name = ""
  _serial_number = ""
  _usb_hub_port = None
  _usb_hub_address = None
  _vendor_id = ""

  def __init__(self, **kwargs):
    for key, value in kwargs.items():
      if not hasattr(self, key):
        raise ValueError("{} is not an arg. Choose from {}".format(
            key, self.get_properties()))
      setattr(self, "_" + key, value)

  def __repr__(self):
    a_dict = {}
    for attr in self.get_properties():
      a_dict[attr] = getattr(self, attr)
    return str(a_dict)

  @classmethod
  def get_properties(cls):
    """Returns all the public properties."""
    return [
        name for name, value in inspect.getmembers(cls)
        if isinstance(value, property) and not name.startswith("_")
    ]

  @property
  def address(self):
    """Address of device, persistent serial port path or adb serial."""
    return self._address

  @address.setter
  def address(self, a_string):
    self._address = a_string

  @property
  def child_addresses(self):
    """If cambrionix, list of addresses that are attached to cambrionix."""
    return self._child_addresses

  @child_addresses.setter
  def child_addresses(self, a_list):
    self._child_addresses = a_list

  @property
  def disk(self):
    """When volume is available, this is the disk link."""
    return self._disk

  @disk.setter
  def disk(self, a_string):
    self._disk = a_string

  @property
  def ftdi_interface(self):
    """Int 0 through 3 representing the ftdi interface."""
    return self._ftdi_interface

  @ftdi_interface.setter
  def ftdi_interface(self, a_number):
    self._ftdi_interface = a_number

  @property
  def manufacturer(self):
    """String that is the manufacturer's name."""
    return self._manufacturer

  @manufacturer.setter
  def manufacturer(self, a_string):
    self._manufacturer = a_string

  @property
  def product_id(self):
    """Product ID in 4 digit form. IE 1201."""
    return self._product_id

  @product_id.setter
  def product_id(self, a_string):
    self._product_id = a_string

  @property
  def product_name(self):
    """Product Name as a string.
    """
    return self._product_name

  @product_name.setter
  def product_name(self, a_string):
    self._product_name = a_string

  @property
  def serial_number(self):
    """Serial number of the connection.

    This could be an FTDI cable or a device.
    """
    return self._serial_number

  @serial_number.setter
  def serial_number(self, a_string):
    self._serial_number = a_string

  @property
  def usb_hub_port(self):
    """Number of the usb port the connection is on, if applicable."""
    return self._usb_hub_port

  @usb_hub_port.setter
  def usb_hub_port(self, a_number):
    self._usb_hub_port = a_number

  @property
  def usb_hub_address(self):
    """Number of the usb port the connection is on, if applicable."""
    return self._usb_hub_address

  @usb_hub_address.setter
  def usb_hub_address(self, a_string):
    self._usb_hub_address = a_string

  @property
  def vendor_id(self):
    """Product ID in 4 digit form. IE 1301."""
    return self._vendor_id

  @vendor_id.setter
  def vendor_id(self, a_string):
    self._vendor_id = a_string



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the TransportProcess class which builds on the SwitchboardProcess base class.

The TransportProcess is responsible for reading and writing bytes to and from
the transport class provided and according to the following assumptions:

    * Transport commands from the main process are received using the
      send_command() method.

    * Raw transport data is (optionally) detokenized.

    * Raw/detokenized data is queued in the expect queue provided.

    * Expect queue can be enabled/disabled by the toggle_expect() method.

    * A host system timestamp is added to every line of data and queued in the
      log queue provided.

    * The host system timestamp will be in the following format:
      "<YYYY-MM-DD hh:mm:ss.ssssss> "

    * Custom log messages received as commands will only be added to the log
      queue between full device log messages.
"""
import queue
import time
import traceback
from typing import Any, Dict, Tuple

from gazoo_device.switchboard import data_framer
from gazoo_device.switchboard import log_process
from gazoo_device.switchboard import switchboard_process
from gazoo_device.switchboard import transport_properties as props
from gazoo_device.utility import multiprocessing_utils

CMD_TRANSPORT_CALL = "TRANSPORT_CALL"
CMD_TRANSPORT_CLOSE = "TRANSPORT_CLOSE"
CMD_TRANSPORT_OPEN = "TRANSPORT_OPEN"
CMD_TRANSPORT_WRITE = "TRANSPORT_WRITE"
PARTIAL_LINE_TIMEOUT = 0.1  # time in seconds before publishing partial lines

_MAX_WRITE_BYTES = 32
_MAX_READ_BYTES = 11520  # 115200 / 10
_READ_TIMEOUT = 0.01  # ((115200 / 10) / 100ms) = ~115 bytes per 10ms read
_ALL_VALID_COMMANDS = (
    CMD_TRANSPORT_CLOSE,
    CMD_TRANSPORT_OPEN,
    CMD_TRANSPORT_WRITE,
    CMD_TRANSPORT_CALL
)


def _enqueue_command_writes(write_queue,
                            command,
                            max_write_bytes=_MAX_WRITE_BYTES):
  """Splits command into max_write_bytes chunks and puts them into queue provided.

  Args:
      write_queue (Queue): to put split command into
      command (str): bytes to be written to device
      max_write_bytes (int): chunks to split command into.

  Note: Splitting commands into smaller write chunks avoids overloading the
    input buffer of the device.
  """

  command_split_indexes = list(range(0, len(command),
                                     max_write_bytes)) + [len(command)]
  for index in range(1, len(command_split_indexes)):
    start_index = command_split_indexes[index - 1]
    end_index = command_split_indexes[index]
    write_queue.put(command[start_index:end_index])


class TransportProcess(switchboard_process.SwitchboardProcess):
  """A process which manages (writes to and reads from) a transport instance.

  Attributes:
    transport: the transport instance used by this transport process.
  """

  def __init__(self,
               device_name,
               exception_queue,
               command_queue,
               log_queue,
               transport,
               call_result_queue,
               raw_data_queue=None,
               raw_data_id=0,
               framer=None,
               partial_line_timeout=PARTIAL_LINE_TIMEOUT,
               read_timeout=_READ_TIMEOUT,
               max_read_bytes=_MAX_READ_BYTES,
               max_write_bytes=_MAX_WRITE_BYTES):
    """Initialize TransportProcess with the arguments provided.

    Args:
      device_name (str): name of device using this transport
      exception_queue (Queue): to use for reporting exception traceback
        message from subprocess
      command_queue (Queue): to receive commands into
      log_queue (Queue): to write each log line with host stamp added
      transport (Transport): to use to receive and send raw data
      call_result_queue (Queue): to write transport call responses to.
      raw_data_queue (Queue): to put raw (if applicable, detokenized) data
        into when enabled.
      raw_data_id (int): unique identifier for data published by this
        transport process to the raw_data_queue.
      framer (DataFramer): to use to frame raw data into partial and
        complete lines.
      partial_line_timeout (float): time in seconds to wait before adding
        partial lines to raw_data_queue and log_queue.
      read_timeout (float): time to wait in seconds for transport reads.
      max_read_bytes (int): to attempt to read on each transport read
        call.
      max_write_bytes (int): to attempt to write on each transport write
        call.
    """
    process_name = "{}-Transport{}".format(device_name, raw_data_id)
    super(TransportProcess, self).__init__(
        device_name,
        process_name,
        exception_queue,
        command_queue,
        valid_commands=_ALL_VALID_COMMANDS)
    self._buffered_unicode = u""
    self._framer = framer or data_framer.NewlineFramer()
    self._log_queue = log_queue
    self._max_read_bytes = max_read_bytes
    self._max_write_bytes = max_write_bytes
    self._partial_line_timeout = partial_line_timeout
    self._partial_log_time = time.time()
    self._pending_writes = None
    self._raw_data_enabled = multiprocessing_utils.get_context().Event()
    self._call_result_queue = call_result_queue
    self._raw_data_id = raw_data_id
    self._raw_data_queue = raw_data_queue
    self._read_timeout = read_timeout
    self._transport_open = multiprocessing_utils.get_context().Event()
    self.transport = transport

  def get_raw_data(self, timeout=None):
    """Returns raw data message from optional raw data queue.

    Args:
        timeout (float): time to wait in seconds for incoming message to
          arrive.

    Raises:
        RuntimeError: if optional expect_queue was not provided when class
                      was initialized

    Returns:
        tuple: A tuple containing the raw_data_id and raw data received from
               the transport device when enabled or None if no data was
               available within timeout specified.
    """
    if self._raw_data_queue is None:
      raise RuntimeError("Device {} can't retrieve raw data. "
                         "No queue provided".format(self.device_name))

    return switchboard_process.get_message(
        self._raw_data_queue, timeout=timeout)

  def is_open(self):
    """Returns True if transport is currently open.

    Returns:
        bool: A boolean indicating transport is currently open.
    """
    return self._transport_open.is_set()

  def raw_data_enabled(self):
    """Returns raw data streaming state.

    Returns:
        bool: A boolean indicating raw data streaming is enabled
    """
    return self._raw_data_enabled.is_set()

  def toggle_raw_data(self):
    """Toggles raw data queue to be enabled/disabled.

    Raises:
        RuntimeError: if optional raw_data_queue was not provided when class
                      was initialized
    """
    if self._raw_data_queue is None:
      raise RuntimeError("Device {} can't enable raw data output. "
                         "No queue provided".format(self.device_name))

    if self._raw_data_enabled.is_set():
      self._raw_data_enabled.clear()
    else:
      self._raw_data_enabled.set()

  def _close_transport(self):
    if self.transport:
      self.transport.close()
      self._transport_open.clear()

  def _do_work(self):
    """Performs transport work and feeding log messages to log_queue.

    Returns:
        bool: always returns True
    """
    command_message = switchboard_process.get_message(
        self._command_queue, timeout=0)
    if command_message:
      self._process_command_message(command_message)
    if self.transport.is_open():
      self._transport_open.set()
      self._transport_write()
      self._transport_read()
    else:
      # Mismatch between self._transport_open and self.transport.is_open
      # means that the transport closed unexpectedly on its own.
      closed_unexpectedly = self._transport_open.is_set()
      can_reopen = self.transport.get_property(props.AUTO_REOPEN, False)
      self._transport_open.clear()

      if closed_unexpectedly:
        self.transport.close()  # Clean up transport resources.
      if closed_unexpectedly and can_reopen:
        self._open_transport()
      else:
        time.sleep(self._read_timeout)
    return True

  def _is_line_published(self, line):
    elapsed_time = time.time() - self._partial_log_time
    if line and (line[-1] == "\n" or elapsed_time > self._partial_line_timeout):
      # NEP-3223 deal with partial line being finished by subsequent line
      if line[-1] == "\n" and self._buffered_unicode:
        self._publish_line(self._buffered_unicode + line)
        self._buffered_unicode = u""
      else:
        self._publish_line(line)
      return True
    return False

  def _open_transport(self):
    self.transport.open()
    self._transport_open.set()

  def _post_run_hook(self):
    """Close transport at end of process."""
    self._close_transport()

  def _pre_run_hook(self):
    """Setup variables and open transport at start of process.

    Returns:
        bool: always returns True
    """
    self._pending_writes = queue.Queue()
    if self.transport.get_property(props.OPEN_ON_START, True):
      self._open_transport()
    self._partial_log_time = time.time()
    return True

  def _process_command_message(self, command_message):
    """Processes command messages which are received as a (command, data) tuple.

    Args:
        command_message (tuple): as received from command queue

    Raises:
        RuntimeError: if command in command_message is unknown.
        ValueError: if data doesn't contain a 2-tuple for CMD_TRANSPORT_SET.
        AttributeError: if the transport doesn't support the called command
        (eg. calling
            CMD_TRANSPORT_SERIAL_FLUSH on a non-serial transport.
    """
    command, data = command_message
    if CMD_TRANSPORT_CLOSE == command:
      self._close_transport()
    elif CMD_TRANSPORT_OPEN == command:
      self._open_transport()
    elif CMD_TRANSPORT_WRITE == command:
      _enqueue_command_writes(
          self._pending_writes, data, max_write_bytes=self._max_write_bytes)
    elif CMD_TRANSPORT_CALL == command:
      self._transport_call(data)
    else:
      raise RuntimeError("Device {} received an unknown command {}.".format(
          self.device_name, command))

  def _publish_line(self, line):
    if self._raw_data_enabled.is_set():
      switchboard_process.put_message(
          self._raw_data_queue, (self._raw_data_id, line), timeout=0)
    log_process.log_message(self._log_queue, line, self._raw_data_id)

  def _transport_call(self, data: Tuple[str,
                                        Tuple[Any],
                                        Dict[str, Any]]) -> None:
    """Calls the transport method and puts result into call_result_queue."""
    try:
      method_name, method_args, method_kwargs = data
      method = getattr(self.transport, method_name)
      return_value = method(*method_args, **method_kwargs)
      success = True
    except Exception:  # pylint: disable=broad-except
      return_value = traceback.format_exc()
      success = False
    self._call_result_queue.put((success, return_value))

  def _transport_read(self):
    """Reads and processing incoming bytes from transport."""

    bytes_in = self.transport.read(
        size=self._max_read_bytes, timeout=self._read_timeout)
    if bytes_in:
      if isinstance(bytes_in, bytes):
        unicode_in = bytes_in.decode("utf-8", "replace")
      else:
        unicode_in = bytes_in
      unicode_in = self._buffered_unicode + unicode_in
      buffered_len = len(self._buffered_unicode)
      self._buffered_unicode = u""
      # Can't use splitlines here because some devices
      # emit multiple line return characters without corresponding line
      # feed characters which prevents desired line splitting behavior.
      for line in self._framer.get_lines(unicode_in, begin=buffered_len):
        self._partial_log_time = time.time()
        if not self._is_line_published(line):
          self._buffered_unicode += line
    elif self._buffered_unicode:
      if self._is_line_published(self._buffered_unicode):
        self._buffered_unicode = u""

  def _transport_write(self):
    """Writes previously split commands into transport."""
    if not self._pending_writes.empty():
      self.transport.write(self._pending_writes.get())



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pigweed RPC transport class."""
import fcntl
import importlib
import queue
import select
import threading
from typing import Any, Collection, Optional, Tuple
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.switchboard.transports import transport_base
from gazoo_device.utility import pwrpc_utils
import serial

# TODO(b/205665429): Consolidate Pigweed import paths.
try:
  # pylint: disable=g-import-not-at-top
  # pytype: disable=import-error
  from pigweed.pw_rpc import client
  from pigweed.pw_rpc import callback_client
  from pigweed.pw_hdlc import rpc
  from pigweed.pw_hdlc import decode
  from pigweed.pw_protobuf_compiler import python_protos
  # pytype: enable=import-error
except ImportError:
  # pylint: disable=g-import-not-at-top
  # pytype: disable=import-error
  # PyPI Pigweed import paths.
  from pw_rpc import client
  from pw_rpc import callback_client
  from pw_hdlc import rpc
  from pw_hdlc import decode
  from pw_protobuf_compiler import python_protos
  # pytype: enable=import-error

_STDOUT_ADDRESS = 1
_DEFAULT_ADDRESS = ord("R")
_JOIN_TIMEOUT_SEC = 1  # seconds
_SELECT_TIMEOUT_SEC = 0.1  # seconds
_SERIAL_TIMEOUT_SEC = 0.01  # seconds
_RPC_CALLBACK_TIMEOUT_SEC = 1  # seconds
logger = gdm_logger.get_logger()


def _is_primitive(var: Any) -> bool:
  """Checks if the given variable is of a primitive type."""
  primitive = (bool, bytes, float, int, str)
  return isinstance(var, primitive)


def _serialize(payload: Any) -> Any:
  """Serializes the given payload.

  Args:
    payload: The payload to serialize. Note that the payload type will be either
      primitive type, protobuf type, or the list type of the above combination.

  Returns:
    Serialized payload.
  """
  if _is_primitive(payload):
    return payload
  if not isinstance(payload, list):
    return payload.SerializeToString()
  return [_serialize(content) for content in payload]


class PwHdlcRpcClient:
  """Pigweed HDLC RPC Client.

  Adapted from
  https://pigweed.googlesource.com/pigweed/pigweed/+/refs/heads/master/pw_hdlc/py/pw_hdlc/rpc.py.
  """

  def __init__(self,
               serial_instance: serial.Serial,
               protobuf_import_paths: Collection[str]):
    """Creates an RPC client configured to communicate using HDLC.

    Args:
      serial_instance: Serial interface instance.
      protobuf_import_paths: Import paths of the RPC proto modules.
    """
    self._serial = serial_instance
    self._protobuf_import_paths = protobuf_import_paths
    self._client = None
    self._stop_event = None
    self._worker = None
    self.log_queue = None

  def is_alive(self) -> bool:
    """Returns true if the worker thread has started."""
    return self._worker is not None and self._worker.is_alive()

  def start(self) -> None:
    """Creates and starts the worker thread if it hasn't been created."""
    if self._worker is None:
      self._stop_event = threading.Event()
      self.log_queue = queue.Queue()
      proto_modules = [importlib.import_module(import_path)
                       for import_path in self._protobuf_import_paths]
      # Load compiled proto modules instead of the raw protos to avoid a
      # "protoc" dependency.
      protos = python_protos.Library.from_paths(proto_modules)
      client_impl = callback_client.Impl(
          default_unary_timeout_s=_RPC_CALLBACK_TIMEOUT_SEC,
          default_stream_timeout_s=_RPC_CALLBACK_TIMEOUT_SEC)
      channels = rpc.default_channels(self._serial.write)
      self._client = client.Client.from_modules(client_impl,
                                                channels,
                                                protos.modules())
      self._worker = threading.Thread(target=self.read_and_process_data)
      self._worker.start()

  def close(self) -> None:
    """Sets the threading event and joins the worker thread."""
    if self._worker is not None:
      self._stop_event.set()
      self._worker.join(timeout=_JOIN_TIMEOUT_SEC)
      if self._worker.is_alive():
        raise errors.DeviceError(
            f"The child thread failed to join after {_JOIN_TIMEOUT_SEC} seconds"
            )
      self._worker = None
      self._client = None
      self._stop_event = None
      self.log_queue = None

  def rpcs(self, channel_id: Optional[int] = None) -> Any:
    """Returns object for accessing services on the specified channel.

    Args:
      channel_id: None or RPC channel id.

    Returns:
      RPC instance over the specificed channel.
    """
    if channel_id is None:
      return next(iter(self._client.channels())).rpcs
    return self._client.channel(channel_id).rpcs

  def read_and_process_data(self) -> None:
    """Continuously reads and handles HDLC frames."""
    decoder = decode.FrameDecoder()
    while not self._stop_event.is_set():
      readfds, _, _ = select.select([self._serial], [], [], _SELECT_TIMEOUT_SEC)
      for fd in readfds:
        try:
          data = fd.read()
        except Exception:  # pylint: disable=broad-except
          logger.exception(
              "Exception occurred when reading in PwHdlcRpcClient thread.")
          data = None
        if data:
          for frame in decoder.process_valid_frames(data):
            self._handle_frame(frame)

  def _handle_rpc_packet(self, frame: Any) -> None:
    """Handler for processing HDLC frame.

    Args:
      frame: HDLC frame packet.
    """
    if not self._client.process_packet(frame.data):
      logger.error(f"Packet not handled by RPC client: {frame.data}")

  def _push_to_log_queue(self, frame: Any) -> None:
    """Pushes the HDLC log in frame into the log queue.

    Args:
      frame: HDLC frame packet.
    """
    self.log_queue.put(frame.data + b"\n")

  def _handle_frame(self, frame: Any) -> None:
    """Private method for processing HDLC frame.

    Args:
      frame: HDLC frame packet.
    """
    try:
      if not frame.ok():
        return
      if frame.address == _DEFAULT_ADDRESS:
        self._handle_rpc_packet(frame)
      elif frame.address == _STDOUT_ADDRESS:
        self._push_to_log_queue(frame)
      else:
        logger.warning(f"Unhandled frame for address {frame.address}: {frame}")
    except Exception:  # pylint: disable=broad-except
      logger.exception("Exception occurred in PwHdlcRpcClient.")


class PigweedRPCTransport(transport_base.TransportBase):
  """Performs transport communication using the Pigweed RPC to end devices."""

  def __init__(self,
               comms_address: str,
               protobuf_import_paths: Collection[str],
               baudrate: int,
               auto_reopen: bool = True,
               open_on_start: bool = True):
    """Initializes a PigweedRPCTransport instance.

    Args:
      comms_address: Serial port path on the host.
      protobuf_import_paths: Module import paths of the compiled device
        communication protobufs. For example,
        ["gazoo_device.protos.device_service_pb2"].
        Proto module objects are not serializable, so their import paths are
        stored instead. Transport processes import the proto modules from these
        paths.
      baudrate: Serial baud rate.
      auto_reopen: Whether to automatically reopen the transport if it closes
        unexpectedly.
      open_on_start: Whether to open the transport during TransportProcess
        start.
    """
    super().__init__(
        auto_reopen=auto_reopen,
        open_on_start=open_on_start)
    self.comms_address = comms_address
    self._serial = serial.Serial()
    self._serial.port = comms_address
    self._serial.baudrate = baudrate
    self._serial.timeout = _SERIAL_TIMEOUT_SEC
    self._hdlc_client = PwHdlcRpcClient(self._serial, protobuf_import_paths)

  def is_open(self) -> bool:
    """Returns True if the PwRPC transport is connected to the target.

    Returns:
      True if transport is open, False otherwise.
    """
    return self._serial.isOpen() and self._hdlc_client.is_alive()

  def _close(self) -> None:
    """Closes the PwRPC transport."""
    self._hdlc_client.close()
    fcntl.flock(self._serial.fileno(), fcntl.LOCK_UN)
    self._serial.close()

  def _open(self) -> None:
    """Opens the PwRPC transport."""
    self._serial.open()
    fd = self._serial.fileno()
    flags = fcntl.fcntl(fd, fcntl.F_GETFD)
    fcntl.fcntl(fd, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)
    self._hdlc_client.start()

  def _read(self, size: int, timeout: float) -> bytes:
    """Returns Pigweed log from the HDLC channel 1.

    Args:
      size: Not used.
      timeout: Maximum seconds to wait to read bytes.

    Returns:
      bytes read from transport or None if no bytes were read
    """
    # Retrieving logs from queue doesn't support size configuration.
    del size  # not used
    try:
      return self._hdlc_client.log_queue.get(timeout=timeout)
    except queue.Empty:
      return b""

  def _write(self, data: str, timeout: Optional[float] = None) -> int:
    """Dummy method for Pigweed RPC.

    Declared to pass the inheritance check of TransportBase.

    Args:
      data: Not used.
      timeout: Not used.

    Returns:
      Always returns 0.
    """
    del data  # not used
    del timeout  # not used
    return 0

  def rpc(self,
          service_name: str,
          event_name: str,
          **kwargs: Any) -> Tuple[bool, Optional[bytes]]:
    """RPC call to the Matter endpoint with given service and event name.

    Args:
      service_name: PwRPC service name.
      event_name: Event name in the given service instance.
      **kwargs: Arguments for the event method.

    Returns:
      (RPC ack value, RPC encoded payload in bytes)
    """
    if not self._hdlc_client.is_alive():
      return False, None

    client_channel = self._hdlc_client.rpcs().chip.rpc
    service = getattr(client_channel, service_name)
    event = getattr(service, event_name)
    kwargs = {
        param_name:
        param.decode() if isinstance(param, pwrpc_utils.PigweedProtoState)
        else param for param_name, param in kwargs.items()}
    ack, payload = event(**kwargs)
    return ack.ok(), _serialize(payload)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Capability base class.

All capabilities must inherit from this class.

Serves two purposes:
- Enforce all capabilities to have a device_name attribute.
  This is a requirement for informative log and error messages in multi-device
  testing.
- Dynamically identify all capabilities supported by GDM.
  This is accomplished by scanning the capabilities folder and identifying all
  classes
  which subclass CapabilityBase.
"""
import abc
import functools
from typing import Any, Collection, Dict, Type
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import extensions
from gazoo_device import gdm_logger
from gazoo_device.utility import common_utils

logger = gdm_logger.get_logger()
_CAPABILITY_INTERFACE_SUFFIX = "_base"


class CapabilityBase(abc.ABC):
  """Abstract base class for all capabilities."""

  def __init__(self, device_name):
    """Set the device_name attribute of the capability.

    Args:
      device_name (str): name of the device instance the capability is attached
        to. Used for error and log messages in multi-device tests.
    """
    self._device_name = device_name
    self._healthy = None

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def close(self):
    """Performs cleanup and releases capability resources."""

  @decorators.DynamicProperty
  def healthy(self):
    """Whether or not the capability has passed a health_check."""
    if self._healthy is None:
      try:
        self.health_check()
        self._healthy = True
      except errors.CapabilityNotReadyError:
        self._healthy = False
    return self._healthy

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def health_check(self):
    """Checks capability readiness. Override in derived classes.

    Raises:
      CapabilityNotReadyError: if the capability is not ready.
    """
    self._healthy = True

  @decorators.DynamicProperty
  def health_checked(self) -> bool:
    """Whether the health_check has been run.

    Returns:
      bool representing whether the health check has been run.
    """
    return self._healthy is not None

  @classmethod
  def get_capability_interface(cls):
    """Returns the capability interface class.

    If called on a flavor class, the nearest parent interface will be
    returned: FileTransferScp.get_capability_interface() ->
      <class file_transfer_base.FileTransferBase>.

    If called on an interface class, the same interface class will be
    returned: FileTransferBase.get_capability_interface() ->
      <class file_transfer_base.FileTransferBase>.

    If called on CapabilityBase class, returns CapabilityBase (special case):
      CapabilityBase.get_capability_interface() -> <class CapabilityBase>.

    Returns:
      type: capability interface class of this capability flavor.

    Raises:
      TypeError: if the capability flavor does not inherit from any
        capability interface.
    """
    if cls is CapabilityBase:
      return CapabilityBase
    for parent_class in cls.__mro__:
      if parent_class in extensions.capability_interfaces.values():
        return parent_class
    raise TypeError(
        "Capability {} does not inherit from a capability interface.".format(
            cls))

  @classmethod
  @functools.lru_cache(maxsize=None)
  def get_capability_name(cls) -> str:
    """Returns the name under which capability is accessible in device class.

    This default implementation can be overridden in special cases.
    Examples:
      FileTransferBase.get_capability_name() -> "file_transfer"
      FileTransferScp.get_capability_name() -> "file_transfer"
      ABCDEventsBase.get_capability_name() -> "abcd_events"
    """
    return get_default_capability_name(cls.get_capability_interface())

  @decorators.CapabilityLogDecorator(logger, level=None)
  def validate_required_keys(self,
                             key_list: Collection[str],
                             dictionary: Dict[str, Any],
                             dictionary_name: str):
    """Verify that the required keys are present in the provided dictionary.

    self._device_name must be set before calling this method (by
    CapabilityBase.__init__). All raised errors are converted to DeviceError by
    the log decorator.

    Args:
      key_list: keys that are required in the dictionary.
      dictionary: to check for presence of all required keys.
      dictionary_name: name (or purpose) of the dictionary. Only
        used in the error message. For example, "Commands".

    Raises:
      KeyError: if a required key is missing from the dictionary.
    """
    missing_keys = [key for key in key_list if key not in dictionary]
    if missing_keys:
      raise KeyError(
          "{} failed to create {!r} capability. Dictionary {!r} is missing the "
          "following keys: {}".format(self._device_name,
                                      self.get_capability_name(),
                                      dictionary_name, missing_keys))


def get_default_capability_name(interface: Type[CapabilityBase]) -> str:
  """Generates the name under which a capability is accessible in device class.

  This is the default name generation logic.
  Examples:
    FileTransferBase -> "file_transfer"
    ABCDEventsBase -> "abcd_events"

  Args:
    interface (type): capability interface (a CapabilityBase subclass).

  Returns:
    str: capability name to use in device class capability definitions.

  Raises:
    RuntimeError: if unable to generate the name because the capability
      interface does not follow the standard naming convention
      (<some_capability>Base).
    ValueError: if the interface is CapabilityBase.
  """
  if interface is CapabilityBase:
    raise ValueError("Cannot generate a capability name for the CapabilityBase "
                     "interface.")
  snake_case_name = common_utils.title_to_snake_case(interface.__name__)
  if not snake_case_name.endswith(_CAPABILITY_INTERFACE_SUFFIX):
    raise RuntimeError(
        "Unable to automatically generate the capability name from "
        "capability interface {}. Interface snake case name {!r} does not "
        "end in {!r}. Either rename the interface or override "
        "get_capability_name() in the interface."
        .format(interface, snake_case_name, _CAPABILITY_INTERFACE_SUFFIX))
  return snake_case_name[:-len(_CAPABILITY_INTERFACE_SUFFIX)]
